
/*
Timemachine
Copyright (c) 2006 Technische Universitaet Muenchen,
                   Technische Universitaet Berlin,
                   The Regents of the University of California
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
#include <stdio.h>

#include "Connections.hh"


Connections::Connections(u_int32_t hash_size):
		hq(hash_size),
		num_entries(0),
		subscriptions(0),
max_subscriptions(0) {
	pthread_mutex_init(&lock_mutex, NULL);
}

/*
Connection* Connections::update_get(ConnectionID* c_id)
{
  return hq.update_get(c_id);
}
*/

Connection* Connections::addPkt_get(const struct pcap_pkthdr* header, const u_char* packet) {
	ConnectionID4* c_id=new ConnectionID4(packet);
	//  printf ("allocated ConnectionID %x\n", c_id);

	Connection* c=hq.update_get(c_id);
	if (c==NULL) {
		c=new Connection;
		//    printf ("allocated Connection %x for %x\n", c, c_id);
		assert(hq.insert_or_update(c_id, c));

		num_entries++;
	} else {
		//    printf ("deallocated ConnectionID %x\n", c_id);
		delete c_id;
	}
	c->addPkt(header, packet);

	return c;
}

Connection* Connections::newConn_get(ConnectionID4 *c_id) {
	Connection* c;
	c=new Connection;
	assert(hq.insert_or_update(c_id, c));
	num_entries++;
	return c;
}

Connection* Connections::getCopy(ConnectionID4 *c_id) {
	lock();
	Connection *c=get(c_id);
	unlock();
	return c ? new Connection(*c) : NULL;
}


void Connections::printStats() const {
	printStats(stdout);
}
	
void Connections::printStats(FILE *fp) const {
	fprintf(fp, "Connections Object Statistics\n"
		   "%"PRINTF_UINT64" total connections stored\n",
		   getNumEntries()
		  );
}

void Connections::debugPrint() const {
	debugPrint(stderr);
}

void Connections::debugPrint(FILE *fp) const {
	hq_t::const_iterator_t i;
	int c;
	fprintf(fp, "-- top 5:\n");
	for (i=hq.nodes.begin(), c=0;
			i!=hq.nodes.end() && c<5; i++, c++)
		fprintf(fp, "* %s\n%s\n", i->getConstK()->getStr().c_str(),
			   i->v->getStr().c_str());
	fprintf(fp, "-- bottom 5\n");
	for (i=hq.nodes.end(), i--, c=0;
			i!=hq.nodes.begin() && c<5; i--, c++)
		fprintf(fp, "* %s\n%s\n", i->getConstK()->getStr().c_str(),
			   i->v->getStr().c_str());
	//  hq.hash.debugPrint();
}


void Connections::removeOld(tm_time_t threshold) {
	Connection* c;
	unsigned loop_iters = 0;

	// XXX: Remove me 
	assert((int)num_entries == hq.hash.getNumEntries());

	while (!hq.empty() && (c=hq.last()) && (c->getLastTs() < threshold)) {
		hq_t::iterator_t it=hq.lastIterator();
		if (c->getSuspendTimeout()) {
			/* connection should not be timed out */
			/* move to front of queue */
			HashQueueNode<ConnectionID4*, Connection*> *ith = &(*it);
			Hash<ConnectionID4*, HashQueue<ConnectionID4*, Connection*>::iterator_t>::iterator_t hashit = ith->hash_iterator;
			HashNode<ConnectionID4*, HashQueue<ConnectionID4*, Connection*>::iterator_t> *hashnode = &(*(hashit.i));
			//      delete hashnode->k;
			//      hq.eraseInQueue(it);
			//      delete c;
#ifdef DEBUG_CONNECTION_QUEUE_SUSPEND_TIMEOUT
			printf("Connections::removeOld suspended conn\nKEY %s\nVAL %s\n",
				   hashnode->k->getStr().c_str(), c->getStr().c_str());
#endif
			hq.insert_or_update(hashnode->k, c);
			
			/* We must ensure, that we can leave the loop when all entries in the
			 * hash queue have a suspended timeout */
			loop_iters++;
			if (loop_iters > num_entries) {
				// We have checked every entry in the hq. Obviously all of them
				// have a suspended timeout. Break the loop
				break;
			}
		} else {
			/* remove timed out connection */
			HashQueueNode<ConnectionID4*, Connection*> *ith = &(*it);
			Hash<ConnectionID4*, HashQueue<ConnectionID4*, Connection*>::iterator_t>::iterator_t hashit = ith->hash_iterator;
			HashNode<ConnectionID4*, HashQueue<ConnectionID4*, Connection*>::iterator_t> *hashnode = &(*(hashit.i));
			//printf ("deallocated ConnectionID %x\n", &ith);
			//    printf ("deallocated ConnectionID %x\n", hashnode->k);
			//delete (*(((*it)->hash_iterator).i))->k;
			delete hashnode->k;
			hq.eraseInQueue(it);
			//    delete c->getKey();
			//printf ("deallocated Connection %x\n", c);

			/* delete possible connection */
			subscriptions -= c->deleteSubscription();
			delete c;

			num_entries--;
		}
	}
}


void Connections::lock() {
	//  while (lock_sem) { sched_yield(); }
	//  lock_sem++;
	pthread_mutex_lock(&lock_mutex);
}

void Connections::unlock() {
	//  lock_sem--;
	pthread_mutex_unlock(&lock_mutex);
}

