/*
Timemachine
Copyright (c) 2006 Technische Universitaet Muenchen,
                   Technische Universitaet Berlin,
                   The Regents of the University of California
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// $Id: Connections.hh 104 2006-11-10 00:46:14Z gregor $

#ifndef CONNECTIONS_HH
#define CONNECTIONS_HH

#include <pthread.h>

#include "Queue.hh"
#include "Connection.hh"


class Connections {
public:
	Connections(u_int32_t hash_size);
	Connection* addPkt_get(const struct pcap_pkthdr*, const u_char*);
	Connection* newConn_get(ConnectionID4*);
	Connection* get(ConnectionID4 *c_id) {
		return hq.lookup(c_id);
	}
	Connection* getCopy(ConnectionID4 *c_id);
	void printStats() const;
	void printStats(FILE *fp) const;
	void debugPrint() const;
	void debugPrint(FILE *fp) const;
	void removeOld(tm_time_t threshold);
	u_int64_t getNumEntries() const {
		return num_entries;
	}
	void lock();
	void unlock();
	void setMaxSubscriptions(int i) {
		max_subscriptions=i;
	}
	int getSubscriptions() {
		return subscriptions;
	}
	int getMaxSubscriptions() {
		return max_subscriptions;
	}
	void increaseSubscriptions() {
		subscriptions++;
	}
protected:
	typedef HashQueue<ConnectionID4*, Connection*> hq_t;
	hq_t hq;
	unsigned num_entries;
	//  u_int32_t lock_sem;
	pthread_mutex_t lock_mutex;
	int subscriptions;
	int max_subscriptions;

};

#endif
