/*
Timemachine
Copyright (c) 2006 Technische Universitaet Muenchen,
                   Technische Universitaet Berlin,
                   The Regents of the University of California
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the names of the copyright owners nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/* main.cc
 * $Id: main.cc 107 2006-11-11 03:11:40Z gregor $
 *
 */

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <pcap.h>
#include <sys/socket.h>
#include <net/ethernet.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
//#include <netinet/tcp.h>
//#include <netinet/udp.h>
#include <arpa/inet.h>
#include <signal.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#define _REENTRANT
#include <readline/readline.h>
#include <readline/history.h>
#include <unistd.h>
#include <netdb.h>
#include <sys/resource.h>
#ifdef __linux
#include <malloc.h>
#define USE_MALLINFO
#endif
#include <sys/ioctl.h>

#ifdef HAVE_LIBBROCCOLI
#include <broccoli.h>
#endif

#ifdef HAVE_LIBPCAPNAV
#include <pcapnav.h>
#endif

#include "conf.h"
#include "types.h"

#include "Storage.hh"
#include "Log.hh"
#include "Query.hh"
#include "Index.hh"


/***************************************************************************
 * global variables
 */

// FIXME: I want a header file
void *
rmtconsole_listen_thread(void *arg);

char errbuf[PCAP_ERRBUF_SIZE]="";

// unsigned int smallest=UINT_MAX, largest=0;
u_int64_t tot_bytes=0, stats_last_call_tot_bytes=0, tot_pkt_cnt=0;
struct timeval stats_last_call_ts;

Storage* storage=NULL;


#define HOSTNAME_MAXLEN 32
char hostname[HOSTNAME_MAXLEN];

LogFile* log_file = NULL;

pthread_t
	rmtconsole_listen_thread_tid,
	cli_console_thread_tid,
	statisticslog_thread_tid,
	index_aggregation_thread_tid,
	main_thread_tid
	;




/* Got some strange effects with signals being received from multiple
 * threads. Hope this mutex helps
 */
//static pthread_mutex_t sigmutex; 

#ifdef HAVE_LIBBROCCOLI
BroConn *bc=NULL;
pthread_mutex_t bc_mutex;
pthread_t broccoli_thread_tid;
int broccoli_recv_q=0;
int broccoli_recv_q_peak=0;
#endif



/***************************************************************************
 * configuration parameters
 */

StorageConfig storageConf;
int conf_main_log_interval=60;
char* conf_main_workdir="./";
char* conf_main_indexdir="./";
char* conf_main_queryfiledir="./";
char* conf_main_logfile_name="tm.log";
char* conf_main_bro_connect_str=NULL;
int conf_main_console=0;
unsigned short conf_main_rmtconsole_port = 42042;
struct in_addr conf_main_rmtconsole_listen_addr; // defualt value is set in main!
int conf_main_rmtconsole = 0;




/***************************************************************************
 * output pcap statistics
 */
void print_stats(FILE *outfp) {
	struct pcap_stat ps;
	storage->getPcapStats(&ps);
	fprintf(outfp, "%u\tpkts received\t[pcap_stats]\n"
		   "%u\tpkts dropped by kernel\t[pcap_stats]\n"
		   "%.2f\t%% dropped/received\t[pcap_stats]\n"
		   "%"PRINTF_UINT64"\tpkts total received by capture application\n"
		   "%"PRINTF_UINT64"\tbytes total\n",
		   ps.ps_recv, ps.ps_drop,
		   (float)ps.ps_drop/ps.ps_recv*100,
		   tot_pkt_cnt, tot_bytes);
	if (stats_last_call_ts.tv_sec) {
		struct timeval now;
		gettimeofday(&now, NULL);
		double dt=now.tv_sec-stats_last_call_ts.tv_sec
				  + (now.tv_usec-stats_last_call_ts.tv_usec)/1e6;
		/* ((now.tv_sec-stats_last_call_ts.tv_sec)*1e6
		   +now.tv_usec-stats_last_call_ts.tv_usec)/1e6; */
		fprintf(outfp, "%.1f\tMbit/s avg since last stats call (%.1fs ago)\n",
			   (tot_bytes-stats_last_call_tot_bytes)/dt/125000, dt);
	}

	gettimeofday(&stats_last_call_ts, NULL);
	stats_last_call_tot_bytes=tot_bytes;

	storage->debugPrint(outfp);

	storage->getConns().printStats(outfp); // number of connection entries

	//  storage->conns.debugPrint(); // top and bottom 5

	fprintf(outfp, "getOldestTimestampMem()=%f\n", storage->getOldestTimestampMem());
	// Formatproblem: printf("tot_num_queries %lu\n", storage->getTotNumQueries());
	// 
	// Valgrind hat gemeckert!
	//printf("avg_queries_duration %f us\n",
		   //(double)storage->getTotQueriesDuration()/
		   //storage->getTotNumQueries());
}





/***************************************************************************
 * end program
 */

void
tmexit() {
	
	// might be not yet initialized, if exit() is called shortly after startup
///	if (storage->isStarted()) {
		printf("print_stats...\n");
		print_stats(stdout);
//	}
//	
	#ifdef HAVE_LIBBROCCOLI
	// Cancel broccoli thread 
	fprintf(stderr, "Canceling broccoli thread\n");
	pthread_cancel(broccoli_thread_tid);
	fprintf(stderr, "Joining broccoli thread\n");
	pthread_join(broccoli_thread_tid, NULL);
	fprintf(stderr, "Broccoli thread i DEAD.\n");
	#endif
	if (conf_main_rmtconsole) {
		// Cancel rmtconsole_listen thread 
		fprintf(stderr, "Canceling rmtconsole_listen thread\n");
		pthread_cancel(rmtconsole_listen_thread_tid);
		fprintf(stderr, "Joining rmtconsole_listen thread\n");
		pthread_join(rmtconsole_listen_thread_tid, NULL);
		fprintf(stderr, "rmtconsole listen thread i DEAD.\n");
	}
	
	fprintf(stderr, "Canceling stats thread\n");
	pthread_cancel(statisticslog_thread_tid);
	fprintf(stderr, "Joining stats thread\n");
	pthread_join(statisticslog_thread_tid, NULL);
	fprintf(stderr, "stats thread i DEAD.\n");

	// Cancel aggregation thread 
	fprintf(stderr, "Canceling aggreagation thread\n");
	pthread_cancel(index_aggregation_thread_tid);
	fprintf(stderr, "Joining aggreagation thread\n");
	pthread_join(index_aggregation_thread_tid, NULL);
	fprintf(stderr, "Aggreagation thread i DEAD.\n");
	//TODO: Cancel all query threads 
	//XXX: Assumption: when a thread waits for lock it is cancel-able

	// when profiling: exit immediately here
	// exit(0);


	// Deleting the storage object will break the pcap loop and thus
	// the capture thread will quit.
	// Deactivating storage will also destroy all indexes and will 
	// thus cancel all index maintaining threads
	printf("deactivating storage... ");
	storage->cancelThread();
	delete storage;


	printf("done\n");
	log_file->log("main", "Storage deactivated");

	/*
	printf("deactivating log file...\n");
	LogFile* actual_log_file=log_file;
	log_file=NULL;
	delete actual_log_file;
	*/

	delete log_file;

	printf("final exit...\n");

	cmd_parser_finish();
	exit(0);
	// No need to unlock -- we are DEAD by now
}


/***************************************************************************
 * signal handler 
 */

void sighandler_exit(int sig_num) {
	//FIXME: Calling printf from a signal handler ! Still a problem with thread
	//save libc??
	// do the same for all signals

	//signal(sig_num, SIG_IGN);
	if(pthread_equal(pthread_self(), main_thread_tid)) {
		printf("exiting due to signal %d...\n", sig_num);
		//pthread_mutex_lock(&sigmutex);
		tmexit();
	}
	else 
		printf("[[[ Nicht die Mama ]]]\n");
} // catch_int()




/***************************************************************************
 * CLI console thread
 */

void *cli_console_thread(void *arg) {
	/*  char buf[8192]; */
	char *line, *lastline=NULL;
	//  char *opt[2];


	char prompt[64];
	snprintf(prompt, 63, "tm@%s# ", hostname);

	printf("CLI console thread started\n");
	log_file->log("main", "CLI console thread started");

	do {
		line=readline(prompt);
		if (line !=NULL && strlen(line)>0) {
			if (!lastline || strcmp(lastline, line)) 
				add_history(line);
			if (lastline) 
				free(lastline);
			lastline=strdup(line);
			parse_cmd(line, stdout, storage);
			free(line);
		}
	} while (line != NULL);

	printf("CLI console thread exiting\n");
	log_file->log("main", "CLI console thread exiting");

	return(0);
} /* cli_console_thread */


#ifdef HAVE_LIBBROCCOLI


/***************************************************************************
 * callback handler for events received by broccoli
 */

void cmd_callback(BroConn* bc, void* userdata, BroString* bro_str) {
	// duplicate BroString, convert to a NULL-terminated char*
	char *c_str=(char *)malloc(bro_str->str_len+1);
	memcpy(c_str, bro_str->str_val, bro_str->str_len);
	c_str[bro_str->str_len]='\0';
	//  printf("cmd_callback(%p, %p, \"%s\")\n", bc, userdata, c_str);
	//  int bro_queue_len=bro_event_queue_length(bc);
	//  if (bro_queue_len)
	//    printf("bro_event_queue_length() is %d\n", bro_queue_len);
	//
	//FIXME: where to send the output?
	parse_cmd(c_str, stdout, storage);
	free(c_str);
};


/***************************************************************************
 * broccoli server thread
 */

void *broccoli_thread(void *arg) {
	/*
	bro_debug_calltrace = 1;
	bro_debug_messages = 1;
	*/
	pthread_mutex_init(&bc_mutex, NULL);
	pthread_mutex_lock(&bc_mutex);
	if (conf_main_bro_connect_str) {
		bc=bro_conn_new_str(conf_main_bro_connect_str,
							BRO_CFLAG_NONE
							// | BRO_CFLAG_ALWAYS_QUEUE
							// | BRO_CFLAG_SHAREABLE
						   );
		if (!bc) {
			log_file->log("broccoli", "Bro connection to %s failed (could not get broccoli handle); "
						  "terminating Broccoli thread\n",
						  conf_main_bro_connect_str);
			fprintf(stderr, "Bro connection to %s failed (could not get broccoli handle); "
					"terminating Broccoli thread\n",
					conf_main_bro_connect_str);
			return(NULL);
		}

		bro_event_registry_add(bc, "TimeMachine::command",
							   (BroEventFunc)cmd_callback, NULL);
		bro_event_registry_request(bc);

		if (!bro_conn_connect(bc)) {
			log_file->log("broccoli", "Bro connection to %s failed; "
						  "terminating Broccoli thread\n",
						  conf_main_bro_connect_str);
			fprintf(stderr, "Bro connection to %s failed; "
					"terminating Broccoli thread\n",
					conf_main_bro_connect_str);
			return(NULL);
		}
		log_file->log("broccoli", "connected to %s", conf_main_bro_connect_str);

		bro_conn_set_packet_ctxt(bc, storage->getPcapDatalink());

		//    printf("broccoli handle %p\n", bc);

		// select loop
		fd_set rfds;
		int bc_fd=bro_conn_get_fd(bc);
		//    printf("broccoli filedescriptor %d\n", bc_fd);
		//    printf("bc->state %d || bc->state->tx_dead %d || bc->state->rx_dead %d || !bc->bio %d\n", bc->state, bc->state->tx_dead, bc->state->rx_dead, !bc->bio);

		if (bc_fd<0) {
			fprintf(stderr, "bro_conn_get_fd() failed; terminating broccoli thread\n");
			return(NULL);
		}

		FD_ZERO(&rfds);
		FD_SET(bc_fd, &rfds);

		pthread_mutex_unlock(&bc_mutex);

		/* select loop
		 */
		while (1) {
			int sel_rc=select(bc_fd+1, &rfds, NULL, NULL, NULL);
			if (sel_rc<0) {
				//	fprintf(stderr, "Bro connection broke; terminating Broccoli thread\n");
				fprintf(stderr, "Bro connection broke; attempting reconnect\n");
				if (!bro_conn_reconnect(bc)) {
					sleep(1);
					fprintf(stderr, "reconnection attempt failed\n");
				}

				/*
				log_file->log("broccoli",
					      "Bro connection broke; terminating Broccoli thread");
				return(NULL);
				*/
			} // if (select error)
			pthread_mutex_lock(&bc_mutex);
			if (!bro_conn_process_input(bc)) {
				fprintf(stderr, "Bro connection broke; attempting reconnect\n");
				if (!bro_conn_reconnect(bc)) sleep(1);
				/*
				fprintf(stderr, "Bro connection broke; terminating Broccoli thread\n");
				log_file->log("broccoli",
					      "Bro connection broke; terminating Broccoli thread");
				return(NULL);
				*/
			}
			pthread_mutex_unlock(&bc_mutex);
			ioctl(bc_fd, FIONREAD, &broccoli_recv_q);
			if (broccoli_recv_q_peak < broccoli_recv_q)
				broccoli_recv_q_peak = broccoli_recv_q;
		} // while (1)
	} // if (conf_main_bro_connect_str)
	return(NULL);
} // broccoli_thread


/***************************************************************************
 * start broccoli_thread
 */

void start_broccoli_thread() {
	int i=pthread_create(&broccoli_thread_tid, NULL, broccoli_thread, NULL);
	if (i) {
		perror("thread");
		exit(-1);
	}
	log_file->log("broccoli", "broccoli_thread started %d", broccoli_thread_tid);
}


#endif // HAVE_LIBBROCCOLI




/***************************************************************************
 * capture thread
 */
void *index_aggregation_thread(void *arg) {
	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	log_file->log("main", "Index Aggregation Thread Started");
	while (1) {
		sleep(3);
		storage->aggregateIndexFiles();
	}
	return NULL;
}



/***************************************************************************
 * statisticslog thread
 * args:
 *  logging interval: int*
 */

void *statisticslog_thread(void* arg) {
#define LOG_STRING_SIZE 100
	struct timeval last_call_ts;
	struct timeval last_call_ru_utime;
	struct timeval last_call_ru_stime;
	struct pcap_stat ps;

	u_int64_t last_call_tot_bytes=0;
	memset(&last_call_ts, 0, sizeof(last_call_ts));
	memset(&last_call_ru_utime, 0, sizeof(last_call_ru_utime));
	while (log_file) {
		char s[LOG_STRING_SIZE];
		if (storage->getPcapStats(&ps) < 0 ) {
			///TODO:
		}
		snprintf(s, LOG_STRING_SIZE,
				 "%u/%u recvd/dropd P "
				 "(%.2f%%) "
				 "%"PRINTF_UINT64" Pkts, "
				 "%"PRINTF_UINT64" B, ",
				 ps.ps_recv, ps.ps_drop,
				 (float)ps.ps_drop/ps.ps_recv*100,
				 tot_pkt_cnt, tot_bytes
				);

		struct rusage r;
		getrusage(RUSAGE_SELF, &r);

#ifdef USE_MALLINFO
		struct mallinfo mi=mallinfo();
#endif

		double dt=0;
		if (last_call_ts.tv_sec) {
			struct timeval now;
			gettimeofday(&now, NULL);
			dt=now.tv_sec-last_call_ts.tv_sec
			   + (now.tv_usec-last_call_ts.tv_usec)/1e6;
			/* ((now.tv_sec-stats_last_call_ts.tv_sec)*1e6
			+now.tv_usec-stats_last_call_ts.tv_usec)/1e6; */
			char s2[LOG_STRING_SIZE];
			strcpy(s2, s);
			snprintf(s, LOG_STRING_SIZE,
					 "%s %.1f Mbit/s", s2,
					 (tot_bytes-last_call_tot_bytes)/dt/125000
					);
			//      snprintdtrncat(s, s2, LOG_STRING_SIZE);
		}

		gettimeofday(&last_call_ts, NULL);
		last_call_tot_bytes=tot_bytes;

		log_file->log("stats", s);

		storage->logStatsClasses();
//		log_file->log("stats_indexes", storage->getStatsIndexesStr().c_str());
		log_file->log("stats_conns", "%llu conns",
					  storage->getConns().getNumEntries());

//		log_file->log("stats_queries", "%u query subscriptions",
//					  storage->getConns().getSubscriptions());
#ifdef HAVE_LIBBROCCOLI
//		log_file->log("stats_broccoli", "peak receive queue %d bytes",
//					  broccoli_recv_q_peak);
		broccoli_recv_q_peak=0;
#endif
		// http://www.google.com/search?hl=en&lr=&q=getrusage+proc&btnG=Search

		/*
		if (dt) {
		struct timeval ucpu_dt;
		struct timeval scpu_dt;
		timersub(&r.ru_utime, &last_call_ru_utime, &ucpu_dt);
		timersub(&r.ru_stime, &last_call_ru_stime, &scpu_dt);
		log_file->log("stats_rusage", "%.2f%% CPU (%.2f%% user+%.2f%% sys) %ld MAXRSS",
		100*to_tm_time(&ucpu_dt)/dt+100*to_tm_time(&scpu_dt)/dt,
		100*to_tm_time(&ucpu_dt)/dt,
		100*to_tm_time(&scpu_dt)/dt,
		r.ru_maxrss
		);
		*/
/*		log_file->log("stats_rusage", "%.2f s user + %.2f s sys CPU  %ld MAXRSS[kB]",
					  to_tm_time(&r.ru_utime),
					  to_tm_time(&r.ru_stime),
					  r.ru_maxrss/1024
					 );*/
		//    } // if (dt)

#ifdef USE_MALLINFO
		/*
		log_file->log("stats_mallinfo", "%d arena  %d uordblks",
					  mi.arena, mi.uordblks
					 );
		*/
#endif

		memcpy(&last_call_ru_utime, &r.ru_utime, sizeof(last_call_ru_utime));
		memcpy(&last_call_ru_stime, &r.ru_stime, sizeof(last_call_ru_stime));

		sleep(*(int*)arg);
	}
	// log file was closed
	return NULL;
}




/***************************************************************************
 * usage()
 */

void usage() {
	fprintf(stderr, "usage: tm [-i interface] [-r file] [-f filter] [-c config file]\n");
	exit(1);
}



/***************************************************************************
 * main()
 */

int
main(int argc, char** argv) {
	char *conffile="tm.conf";

	struct sigaction exit_action;



	int i;
	inet_aton("127.0.0.1", &conf_main_rmtconsole_listen_addr);

	setvbuf(stdout, NULL, _IONBF, 0);

	main_thread_tid = pthread_self();

	/*******************************************************
	 * read command line arguments
	 */

	int opt;
	while ((opt=getopt(argc, argv, "i:r:f:c:h")) != -1) {
		switch(opt) {
		case 'f':
			storageConf.filter.assign(optarg);
			break;
		case 'i':
			storageConf.device.assign(optarg);
			break;
		case 'r':
			storageConf.readtracefile.assign(optarg);
			break;
		case 'c':
			conffile=strdup(optarg);
			break;
		case 'h':
		default:
			usage();
		}
	}
	argc -= optind;
	argv += optind;



	/***************************************************************************
	 * read and parse config file
	 */
	if (parse_config(conffile, &storageConf)) {
		fprintf(stderr, "config file errors, aborting\n");
		return(1);
	}



	memset(&stats_last_call_ts, 0, sizeof(stats_last_call_ts));

	if (chdir(conf_main_workdir)) {
		fprintf(stderr, "cannot chdir to %s\n", conf_main_workdir);
		return(1);
	}
	free(conf_main_workdir);

	log_file=new LogFile(conf_main_logfile_name);
	log_file->log("main", "timemachine version %s", PACKAGE_VERSION);
	printf("timemachine version %s\n", PACKAGE_VERSION);

	// Blocl all signals. This signal mask is then inherited by spawned 
	// threads. After all threads are spanwed we will unblock again. 
	sigset_t signalSet;
	sigset_t oldSignalSet;
	sigemptyset(&signalSet);
	sigaddset(&signalSet, SIGTERM);
	sigaddset(&signalSet, SIGQUIT);
	sigaddset(&signalSet, SIGINT);
	pthread_sigmask (SIG_BLOCK, &signalSet, &oldSignalSet );
	/*******************************************************
	 * start capture
	 */


	// FIXME: Port this to Storage::Storage
	if (strlen(errbuf)>0) {
		fprintf(stderr, "pcap warning: %s\n", errbuf);
		log_file->log("main", "WARNING: pcap: %s", errbuf);
	}

#ifdef HAVE_LIBPCAPNAV
	pcapnav_init();
#endif

	// Initialise cmd_parser
	cmd_parser_init();

	try {
		storage = new Storage(storageConf);
	} catch (const std::string& str) {
		fprintf(stderr, "%s\n", str.c_str());
		delete log_file;
		exit(1);
	}
	log_file->log("main", "capture started, capture thread");
	printf("capture started\n");

	pthread_create(&index_aggregation_thread_tid, NULL, index_aggregation_thread, NULL);
	/*******************************************************
	 * trap signals
	 */
	exit_action.sa_handler = sighandler_exit;
	exit_action.sa_flags = 0;
	sigemptyset(&(exit_action.sa_mask));
	sigaddset(&(exit_action.sa_mask), SIGTERM);
	sigaddset(&(exit_action.sa_mask), SIGQUIT);
	sigaddset(&(exit_action.sa_mask), SIGINT);
	sigaction(SIGTERM, &exit_action, NULL);
	sigaction(SIGINT, &exit_action, NULL);
	sigaction(SIGQUIT, &exit_action, NULL);
	// when profiling: trap signal USR1
	// signal(SIGUSR1, catch_int);
	//  signal(SIGSEGV, catch_int);


	if (conf_main_rmtconsole) {
		i=pthread_create(&rmtconsole_listen_thread_tid, NULL, rmtconsole_listen_thread, NULL);
		if (i) {
			perror("thread");
			exit(1);
		}
	}


	gethostname(hostname, HOSTNAME_MAXLEN);
	hostname[HOSTNAME_MAXLEN-1]=0;

		i=pthread_create(&cli_console_thread_tid, NULL, cli_console_thread, NULL);
		if (i) {
			perror("thread");
			exit(-1);
		}

#ifdef HAVE_LIBBROCCOLI
		start_broccoli_thread();
#endif

		i=pthread_create(&statisticslog_thread_tid, NULL,
						 statisticslog_thread, &conf_main_log_interval);
		if (i) {
			perror("thread");
			exit(-1);
		};

		// All threads have been spwaned now. Re-enable signal delivery
		// Only this thread will now receive signals. 
		pthread_sigmask (SIG_SETMASK, &oldSignalSet, NULL );

		//   struct sched_param param;
		//   int policy;
		//   param.sched_priority=1;

		//   i=pthread_setschedparam(cli_console_thread_tid, SCHED_FIFO, &param);
		//   if (i) {
		//     perror("thread setschedparam");
		//     exit(-1);
		//   }
		//   i=pthread_getschedparam(cli_console_thread_tid, &policy, &param);
		//   if (i) {
		//     perror("thread getschedparam");
		//     exit(-1);
		//   }
		//   printf(" cli_console_thread schedule priority is %d\n", param.sched_priority);

		if (conf_main_console) {
			pthread_join(cli_console_thread_tid, NULL);
		}
		else {
			// FIXME: Do something usefull here. Maybe pause()
			fprintf(stderr, "NOTE NOTE NOTE NOTE NOTE NOTE NOTE\n");
			fprintf(stderr, "Can't handle it, if we don't hava a consol running\n");
		}

		tmexit();
}

/* A wrapper function to start a thread for each created
 * index object
 */
extern "C" {
void *start_index_thread(void *instance) {
//	pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
	((IndexType *)(instance))->run();
	return NULL;
}
}
